/*************************************************************************
 *                                                                       *
 * GLE - Graphics Layout Engine <http://www.gle-graphics.org/>           *
 *                                                                       *
 * BSD License                                                           *
 *                                                                       *
 * Copyright (c).                                                        *
 * All rights reserved.                                                  *
 *                                                                       *
 * Redistribution and use in source and binary forms, with or without    *
 * modification, are permitted provided that the following conditions    *
 * are met:                                                              *
 *                                                                       *
 *  * Redistributions of source code must retain the above copyright     *
 *  notice, this list of conditions and the following disclaimer.        *
 *                                                                       *
 *  * Redistributions in binary form must reproduce the above            *
 *  copyright notice, this list of conditions and the following          *
 *  disclaimer in the documentation and/or other materials provided      *
 *  with the distribution.                                               *
 *                                                                       *
 *  * Neither the name of GLE nor the names of its contributors may      *
 *  be used to endorse or promote products derived from this software    *
 *  without specific prior written permission.                           *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT      *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT   *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  *
 *                                                                       *
 *************************************************************************/

#ifndef INCLUDE_SourceLine
#define INCLUDE_SourceLine

#include <list>

class _GLESource;

class SourceLine {
public:
	// the global line number for GLE every line of gle source needs its own
	// unique line number thus included files must have this set
	int line_no;

	// the real line number when referencing the source file
	int source_line_no;

	string filename;
	string text;
	_GLESource* source;
public:
	SourceLine();
	SourceLine(const SourceLine& cl);
	~SourceLine();
	SourceLine& operator = (const SourceLine& gr);
	void showPosPointer(int pos, ostream& out);
	const string& getFileName();
	inline void setSource(_GLESource* src) { source = src; }
	inline _GLESource* getSource() { return source; }
};

int operator==(const SourceLine& a , const SourceLine& b);

inline void copy( SourceLine& a , const SourceLine& b) {
    a = b;
}

ostream& operator<<(ostream& out, SourceLine d);

class _GLESource : public std::list<SourceLine> {
protected:
	string m_FName;
public:
	_GLESource();
	~_GLESource();
	inline const string& getFileName() { return m_FName; }
	inline void setFileName(const char* fname) { m_FName = fname; }
	inline void setFileName(const string& fname) { m_FName = fname; }
	void attach(_GLESource& other);
};

typedef std::list<SourceLine>::iterator	_itGLESource;
typedef std::list<SourceLine>::const_iterator _citGLESource;

#endif

