/*************************************************************************
 *                                                                       *
 * GLE - Graphics Layout Engine <http://www.gle-graphics.org/>           *
 *                                                                       *
 * BSD License                                                           *
 *                                                                       *
 * Copyright (c).                                                        *
 * All rights reserved.                                                  *
 *                                                                       *
 * Redistribution and use in source and binary forms, with or without    *
 * modification, are permitted provided that the following conditions    *
 * are met:                                                              *
 *                                                                       *
 *  * Redistributions of source code must retain the above copyright     *
 *  notice, this list of conditions and the following disclaimer.        *
 *                                                                       *
 *  * Redistributions in binary form must reproduce the above            *
 *  copyright notice, this list of conditions and the following          *
 *  disclaimer in the documentation and/or other materials provided      *
 *  with the distribution.                                               *
 *                                                                       *
 *  * Neither the name of GLE nor the names of its contributors may      *
 *  be used to endorse or promote products derived from this software    *
 *  without specific prior written permission.                           *
 *                                                                       *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS   *
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT     *
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR *
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT  *
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, *
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT      *
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, *
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY *
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT   *
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE *
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.  *
 *                                                                       *
 *************************************************************************/

#include "all.h"
#include "SourceLine.h"
#include "gprint.h"
#include "tokens/Tokenizer.h"
#include "core.h"
#include "run.h"

#include "mygraph.h"   /* Prototypes for all the graph routines */
#define true (!false)
#define false 0

int ncvec=0;
void cvec_list(int *pcode);
double cvecx[30],cvecy[30];
double dcvecx[30],dcvecy[30];
void rbezier(double x1, double y1, double x2, double y2, double x3, double y3);

void g_curve(int *pcode) {
	double oox,ooy;
	double oval;
	int ct=2,i;
	double dx1,dy1;

	ncvec = 0;
	cvec_list(pcode);
	dx1 = cvecx[1] - cvecx[0];
	dy1 = cvecy[1] - cvecy[0];
	dcvecx[0] = cvecx[ncvec] - cvecx[ncvec-1];
	dcvecy[0] = cvecy[ncvec] - cvecy[ncvec-1];
	for (i=0;i<=ncvec;i++) {
		cvecx[i] = cvecx[i] - dx1;
		cvecy[i] = cvecy[i] - dy1;
	}
	for (i=1;i<ncvec;i++) {
		dcvecx[i] = (cvecx[i+1] - cvecx[i-1]) * .25;
		dcvecy[i] = (cvecy[i+1] - cvecy[i-1]) * .25;
	}
	for (i=1; i<ncvec-1; i++) {
		rbezier(dcvecx[i],dcvecy[i],dcvecx[i+1],dcvecy[i+1]
			,cvecx[i+1],cvecy[i+1]);
	}
}

void rbezier(double x1, double y1, double x2, double y2, double x3, double y3) {
	double cx,cy;
	g_get_xy(&cx,&cy);
	g_bezier(x1+cx,y1+cy,x3-x2,y3-y2,x3,y3);
}

void cvec_list(int *pcode) {
	char ostr[90];
	int i,cp=0,otyp;
	double cx,cy,x1,y1,x2,y2;
	g_get_xy(&cx,&cy);
	ncvec = 0;
	cvecx[0] = cx;
	cvecy[0] = cy;
	while ( *(pcode + cp++)==111) {
		if (ncvec>27) {gprint("Too many param in curve\n"); return; }
		eval(pcode,&cp,&x1,ostr,&otyp);
		eval(pcode,&cp,&y1,ostr,&otyp);
		cvecx[++ncvec] = x1;
		cvecx[ncvec] = cvecx[ncvec] + cvecx[ncvec-1];
		cvecy[ncvec] = y1;
		cvecy[ncvec] = cvecy[ncvec] + cvecy[ncvec-1];
	}
}

void g_arrowcurve(double x, double y, int arrow, double a1, double a2, double d1, double d2, int can_fill) {
	double cx, cy, x1, y1, x2, y2;
	polar_xy(d1, a1, &x1, &y1);
	polar_xy(d2, a2, &x2, &y2);
	g_get_xy(&cx, &cy);
	g_bezier(cx+x1, cy+y1, x+x2, y+y2, x, y);
	if ((arrow & 1)	!= 0) {
		g_move(cx, cy);
		g_arrow(x1, y1, can_fill);
	}
	if ((arrow & 2)	!= 0) {
		g_move(x, y);
		g_arrow(x2, y2, can_fill);
	}
}
